#include "pose.h"

void Pose::projectWrinkle(pcl::PointCloud<pcl::PointXYZ>::Ptr &wrinkle_cloud,
                          pcl::PointCloud<pcl::PointXYZ>::Ptr &wrinkle_cloud_proj, pcl::ModelCoefficients::Ptr &plane)
{
    pcl::ProjectInliers<pcl::PointXYZ> proj;
    proj.setModelType(pcl::SACMODEL_PLANE);
    proj.setInputCloud(wrinkle_cloud);
    proj.setModelCoefficients(plane);
    proj.filter(*wrinkle_cloud_proj);
}

void Pose::compute(Affine3dVector &transformation_matrix_vector)
{
    std::vector<CoordinateFrameT> frames;
    Eigen::Vector3f normal_dir(m_plane->values[0], m_plane->values[1], m_plane->values[2]);
    normal_dir.normalize();
    Eigen::Affine3d matrix;
    CoordinateFrameT frame;
    std::vector<pcl::ModelCoefficients> lines;
    std::vector<int> skipped;

    for (int i = 0; i < m_wrinkles.size(); i++)
    {
        std::cout << "wrinkle number " << i << ", size: " << m_wrinkles[i]->size() << std::endl;
        pcl::PointCloud<pcl::PointXYZ>::Ptr wrinkle_cloud_proj(new pcl::PointCloud<pcl::PointXYZ>);
        projectWrinkle(m_wrinkles[i], wrinkle_cloud_proj, m_plane);

        m_wrinkles_projected.push_back(wrinkle_cloud_proj);

        pcl::ModelCoefficients line;
        if (!ransacLineFitting(wrinkle_cloud_proj, line))
        {
            skipped.push_back(i);
            continue;
        }

        Eigen::Vector3f wrinkle_dir(line.values[3], line.values[4], line.values[5]);
        Eigen::Vector3f wrinkle_point(line.values[0], line.values[1], line.values[2]);
        wrinkle_dir.normalize();

        translatePoint(m_wrinkles[i], wrinkle_cloud_proj, wrinkle_point);

        matrix = computeTransformation(wrinkle_point, wrinkle_dir, normal_dir);
        std::cout << matrix.matrix() << std::endl;

        frame = createFrameForVisualization(wrinkle_point, wrinkle_dir, normal_dir);

        frames.push_back(frame);
        transformation_matrix_vector.push_back(matrix);
        lines.push_back(line);
    }
    std::cout << "wrinkles size: " << m_wrinkles.size() << std::endl;
    std::cout << "wrinkles proj size: " << m_wrinkles_projected.size() << std::endl;

    visualize(frames, skipped);
}

void Pose::visualize(std::vector<CoordinateFrameT> &frames, std::vector<int> &skipped)
{
    //----------------------------------------------------

    pcl::visualization::PCLVisualizer viz("Cloud Seg Pose Visualizer");
    viz.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity(m_cloud_seg, "intensity");
    viz.addPointCloud<pcl::PointXYZI>(m_cloud_seg, intensity, "m_cloud_seg");
    viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "m_cloud_seg");

    for (int k = 0; k < frames.size(); k++)
    {
        viz.addArrow(frames[k].x, frames[k].o, 1.0f, 0.0f, 0.0f, false, "x_axis" + std::to_string(k));
        viz.addArrow(frames[k].y, frames[k].o, 0.0f, 1.0f, 0.0f, false, "y_axis" + std::to_string(k));
        viz.addArrow(frames[k].z, frames[k].o, 0.0f, 0.0f, 1.0f, false, "z_axis" + std::to_string(k));
    }

    // ------------------------------------------------------------------------------------------------------------------
    pcl::visualization::PCLVisualizer vizS("Source Pose Visualizer");
    vizS.setBackgroundColor(1.0f, 1.0f, 1.0f);
    vizS.addPointCloud<pcl::PointXYZRGB>(m_source, "source");
    vizS.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.0f, 0.7f, 0.0f, "source");

    for (int k = 0; k < frames.size(); k++)
    {
        vizS.addArrow(frames[k].x, frames[k].o, 0.0f, 0.0f, 1.0f, false, "x_axis" + std::to_string(k));
        vizS.addArrow(frames[k].y, frames[k].o, 0.0f, 1.0f, 0.0f, false, "y_axis" + std::to_string(k));
        //vizS.addArrow(frames[k].z, frames[k].o, 1.0f, 0.0f, 0.0f, false, "z_axis" + std::to_string(k));
    }
    for (int i = 0; i < m_wrinkles.size(); i++)
    {
        //if (skipped[0] == i)
        //    continue;

        vizS.addPointCloud<pcl::PointXYZ>(m_wrinkles[i], "m_wrinkles" + std::to_string(i));
        vizS.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 1.0f, 0.0f, 0.0f,
                                              "m_wrinkles" + std::to_string(i));
        vizS.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 6,
                                              "m_wrinkles" + std::to_string(i));
        for (int k = 1; k < m_wrinkles[i]->size(); k++)
        {
            vizS.addLine(m_wrinkles[i]->points[k - 1], m_wrinkles[i]->points[k], 1.0f, 0.0f, 0.0f, "BBBB" + std::to_string(i) + std::to_string(k));
        }
    }
}

Eigen::Affine3d Pose::computeTransformation(Eigen::Vector3f &origin, Eigen::Vector3f &directionX, Eigen::Vector3f &directionZ)
{
    directionX.normalize();
    directionZ.normalize();

    Eigen::VectorXd from_line_x, from_line_z, to_line_x, to_line_z;
    from_line_x.resize(6);
    from_line_z.resize(6);
    to_line_x.resize(6);
    to_line_z.resize(6);

    //Origin
    from_line_x << 0, 0, 0, 1, 0, 0;
    from_line_z << 0, 0, 0, 0, 0, 1;

    to_line_x.head<3>() = origin.cast<double>();
    to_line_x.tail<3>() = directionX.cast<double>();

    to_line_z.head<3>() = origin.cast<double>();
    to_line_z.tail<3>() = directionZ.cast<double>();

    Eigen::Affine3d transformation;
    if (!pcl::transformBetween2CoordinateSystems(from_line_x, from_line_z, to_line_x, to_line_z, transformation))
    {
        PCL_WARN("Transformation not found!\n");
    }
    return transformation;
}

CoordinateFrameT Pose::createFrameForVisualization(Eigen::Vector3f &wrinkle_point,
                                                   Eigen::Vector3f &wrinkle_dir,
                                                   Eigen::Vector3f &normal_dir)
{
    CoordinateFrameT frame;
    frame.o = {wrinkle_point.x(), wrinkle_point.y(), wrinkle_point.z()};

    frame.x = {wrinkle_point.x() + 0.05 * wrinkle_dir.x(),
               wrinkle_point.y() + 0.05 * wrinkle_dir.y(),
               wrinkle_point.z() + 0.05 * wrinkle_dir.z()};

    frame.z = {wrinkle_point.x() + 0.05 * normal_dir.x(),
               wrinkle_point.y() + 0.05 * normal_dir.y(),
               wrinkle_point.z() + 0.05 * normal_dir.z()};

    Eigen::Vector3f third_axis = wrinkle_dir.cross(normal_dir);
    third_axis.normalize();
    frame.y = {wrinkle_point.x() + 0.05 * third_axis.x(),
               wrinkle_point.y() + 0.05 * third_axis.y(),
               wrinkle_point.z() + 0.05 * third_axis.z()};

    return frame;
}

bool Pose::ransacLineFitting(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::ModelCoefficients &line)
{
    pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
    pcl::SACSegmentation<pcl::PointXYZ> seg;

    seg.setOptimizeCoefficients(true);
    seg.setModelType(pcl::SACMODEL_LINE);
    seg.setMethodType(pcl::SAC_RANSAC);
    seg.setDistanceThreshold(0.005);

    seg.setInputCloud(cloud);
    seg.segment(*inliers, line);

    if (inliers->indices.size() == 0)
    {
        std::cerr << "Could not estimate a linear model for the given dataset." << std::endl;
        return false;
    }
    return true;
}

void Pose::translatePoint(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_proj,
                          Eigen::Vector3f &point)
{
    pcl::PointXYZ P = {point.x(), point.y(), point.z()};

    pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
    kdtree.setInputCloud(cloud_proj);
    pcl::Indices indices;
    std::vector<float> sqrDistances;
    int K = 1;

    int index = 0;
    if (kdtree.nearestKSearch(P, K, indices, sqrDistances) > 0)
    {
        //point.z() = cloud->points[indices[0]].z;
        index = indices[0];
    }

    if (index != 0)
    {
        point.x() = cloud->points[index].x;
        point.y() = cloud->points[index].y;
        point.z() = cloud->points[index].z;
    }
}