#include "wrinkle.h"

pcl::PointCloud<pcl::PointXYZI>::Ptr Poly::computeRadiusNeighborhoodCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                                                                          pcl::PointXYZ &searchPoint, float radius)
{
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_extracted(new pcl::PointCloud<pcl::PointXYZI>);

    pcl::PointXYZI searchP;
    searchP.x = searchPoint.x;
    searchP.y = searchPoint.y;
    searchP.z = searchPoint.z;

    pcl::KdTreeFLANN<pcl::PointXYZI> kdtree;
    kdtree.setInputCloud(cloud);
    pcl::Indices indices;
    std::vector<float> sqrDistances;

    if (kdtree.radiusSearch(searchP, radius, indices, sqrDistances) > 0)
    {
        //std::cout << "NN found: " << indices.size() << std::endl;

        pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
        inliers->indices = indices;

        pcl::ExtractIndices<pcl::PointXYZI> extract;
        extract.setInputCloud(cloud);
        extract.setIndices(inliers);
        extract.setNegative(false);
        extract.filter(*cloud_extracted);

        //std::cout << "number of points extracted: " << cloud_extracted->size() << std::endl;
    }
    return cloud_extracted;
}

void Poly::computeDirection(pcl::PointCloud<pcl::PointXYZI>::Ptr cloud, Eigen::Vector3f &centroid, Eigen::Vector3f &direction)
{
    Eigen::Vector4f centroid4f;
    pcl::compute3DCentroid(*cloud, centroid4f);
    Eigen::Matrix3f covariance;
    pcl::computeCovarianceMatrixNormalized(*cloud, centroid4f, covariance);

    Eigen::Vector3f eigenValues;
    Eigen::Matrix3f eigenVectors;
    pcl::eigen33(covariance, eigenVectors, eigenValues);

    std::vector<double> v;
    v.push_back(eigenValues[0]);
    v.push_back(eigenValues[1]);
    v.push_back(eigenValues[2]);

    int maxElementIndex = std::max_element(v.begin(), v.end()) - v.begin();

    direction = eigenVectors.col(maxElementIndex);
    direction.normalize();
    centroid = centroid4f.head<3>();
}

pcl::PointXYZ Poly::computeNewSearchPointForward(Eigen::Vector3f &centroid, Eigen::Vector3f &direction, pcl::PointXYZ &parent, float step)
{
    pcl::PointXYZ point = {0, 0, 0};

    Eigen::Vector3f vec3f = centroid + step * direction;
    Eigen::Vector3f vec1 = vec3f - centroid;
    Eigen::Vector3f vec2 = parent.getVector3fMap() - centroid;
    vec1.normalize();
    vec2.normalize();

    float angle = acos(vec1.dot(vec2));
    //std::cout << angle << std::endl;

    if (angle < M_PI_2)
        vec3f = centroid - step * direction;

    vec1 = vec3f - centroid;
    vec1.normalize();
    angle = acos(vec1.dot(vec2));
    std::cout << angle << std::endl;

    if (fabs(angle - M_PI_2) < 0.3)
        return point;

    point = {vec3f.x(), vec3f.y(), vec3f.z()};
    return point;
}

pcl::PointXYZ Poly::computeNewSearchPointBackward(Eigen::Vector3f &centroid, Eigen::Vector3f &direction, pcl::PointXYZ &parent, float step)
{
    pcl::PointXYZ point = {0, 0, 0};

    Eigen::Vector3f vec3f = centroid - step * direction;
    Eigen::Vector3f vec1 = vec3f - centroid;
    Eigen::Vector3f vec2 = parent.getVector3fMap() - centroid;
    vec1.normalize();
    vec2.normalize();

    float angle = acos(vec1.dot(vec2));
    //std::cout << angle << std::endl;

    if (angle < M_PI_2)
        vec3f = centroid + step * direction;

    vec1 = vec3f - centroid;
    vec1.normalize();
    angle = acos(vec1.dot(vec2));
    std::cout << angle << std::endl;

    if (fabs(angle - M_PI_2) < 0.3)
        return point;

    point = {vec3f.x(), vec3f.y(), vec3f.z()};
    return point;
}

bool Poly::checkSearchPoint(pcl::PointXYZ &searchPoint, float threshold)
{
    if (searchPoint.getVector3fMap().isZero())
        return false;

    pcl::PointXYZI searchP;
    searchP.x = searchPoint.x;
    searchP.y = searchPoint.y;
    searchP.z = searchPoint.z;

    pcl::KdTreeFLANN<pcl::PointXYZI> kdtree;
    kdtree.setInputCloud(m_source);
    pcl::Indices indices;
    std::vector<float> sqrDistances;
    int k = 1;

    if (kdtree.nearestKSearch(searchP, k, indices, sqrDistances) > 0)
    {
        if (sqrt(sqrDistances[0]) > threshold)
        {
            std::cout << "NN found at distance: " << sqrt(sqrDistances[0]) << " m" << std::endl;
            return false;
        }
    }
    return true;
}

void Poly::computePointsInit(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                             PolypointsT &pointsF, PolypointsT &pointsB, float &radius, float &step)
{
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_extracted(new pcl::PointCloud<pcl::PointXYZI>);
    Eigen::Vector3f centroid, direction;
    cloud_extracted = computeRadiusNeighborhoodCloud(cloud, pointsF.search, radius);
    computeDirection(cloud_extracted, centroid, direction);

    pcl::PointXYZ centroidP = {centroid.x(), centroid.y(), centroid.z()};
    pointsF.centroid = centroidP;
    pointsB.centroid = centroidP;
    pointsF.pointParent = centroidP;
    pointsB.pointParent = centroidP;

    pcl::PointXYZ pointF = computeNewSearchPointForward(centroid, direction, pointsF.pointParent, step);
    if (checkSearchPoint(pointF, radius) == 1) // check passed!
        pointsF.pointComputed = pointF;

    pcl::PointXYZ pointB = computeNewSearchPointBackward(centroid, direction, pointsB.pointParent, step);
    if (checkSearchPoint(pointB, radius) == 1) // check passed!
        pointsB.pointComputed = pointB;
}

bool Poly::computeForward(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                          PolypointsT &points, float &radius, float &step)
{
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_extracted(new pcl::PointCloud<pcl::PointXYZI>);
    Eigen::Vector3f centroid, direction;
    cloud_extracted = computeRadiusNeighborhoodCloud(cloud, points.search, radius);
    computeDirection(cloud_extracted, centroid, direction);

    pcl::PointXYZ centroidP = {centroid.x(), centroid.y(), centroid.z()};
    points.centroid = centroidP;

    pcl::PointXYZ P = computeNewSearchPointForward(centroid, direction, points.pointParent, step);
    if (checkSearchPoint(P, radius / 2) == 1) // check passed!
    {
        points.pointComputed = P;
        return true;
    }
    return false;
}

bool Poly::computeBackward(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                           PolypointsT &points, float &radius, float &step)
{
    pcl::PointCloud<pcl::PointXYZI>::Ptr cloud_extracted(new pcl::PointCloud<pcl::PointXYZI>);
    Eigen::Vector3f centroid, direction;
    cloud_extracted = computeRadiusNeighborhoodCloud(cloud, points.search, radius);
    computeDirection(cloud_extracted, centroid, direction);

    pcl::PointXYZ centroidP = {centroid.x(), centroid.y(), centroid.z()};
    points.centroid = centroidP;

    pcl::PointXYZ P = computeNewSearchPointBackward(centroid, direction, points.pointParent, step);
    if (checkSearchPoint(P, radius / 2) == 1) // check passed!
    {
        points.pointComputed = P;
        return true;
    }
    return false;
}

pcl::PointXYZ Poly::initialSearchPoint()
{
    // compute initial search point as the maximum entropy point
    float max = 0;
    int it;
    for (int i = 0; i < m_source->size(); i++)
    {
        if (m_source->points[i].intensity > max)
        {
            max = m_source->points[i].intensity;
            it = i;
        }
    }
    //std::cout << "max entropy: " << max << ", at indices: " << it << std::endl;
    pcl::PointXYZ searchPoint0 = {m_source->points[it].x, m_source->points[it].y, m_source->points[it].z};
    return searchPoint0;
}

pcl::PointXYZ Poly::newSearchPoint(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                                   pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_polyF,
                                   pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_polyB)
{
    std::cout << "source copy initial size: " << cloud->size() << std::endl;

    for (int k = 0; k < cloud_polyF->size(); k++)
    {
        if (cloud->size() == 0)
            break;

        pcl::PointXYZ search = cloud_polyF->points[k];
        pcl::PointXYZI searchI;
        searchI.x = search.x;
        searchI.y = search.y;
        searchI.z = search.z;

        pcl::KdTreeFLANN<pcl::PointXYZI> kdtree;
        kdtree.setInputCloud(cloud);
        pcl::Indices indices;
        std::vector<float> sqrDistances;

        if (kdtree.radiusSearch(searchI, m_radius * 3, indices, sqrDistances) > 0)
        {
            pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
            inliers->indices = indices;

            pcl::ExtractIndices<pcl::PointXYZI> extract;
            extract.setInputCloud(cloud);
            extract.setIndices(inliers);
            extract.setNegative(true);
            extract.filter(*cloud);
        }
        if (cloud->size() == 0)
            break;
    }

    for (int k = 0; k < cloud_polyB->size(); k++)
    {
        if (cloud->size() == 0)
            break;

        pcl::PointXYZ search = cloud_polyB->points[k];
        pcl::PointXYZI searchI;
        searchI.x = search.x;
        searchI.y = search.y;
        searchI.z = search.z;

        pcl::KdTreeFLANN<pcl::PointXYZI> kdtree;
        kdtree.setInputCloud(cloud);
        pcl::Indices indices;
        std::vector<float> sqrDistances;

        if (kdtree.radiusSearch(searchI, m_radius * 3, indices, sqrDistances) > 0)
        {
            pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
            inliers->indices = indices;

            pcl::ExtractIndices<pcl::PointXYZI> extract;
            extract.setInputCloud(cloud);
            extract.setIndices(inliers);
            extract.setNegative(true);
            extract.filter(*cloud);
        }

        if (cloud->size() == 0)
            break;
    }

    std::cout << "source copy size: " << cloud->size() << std::endl;
    pcl::PointXYZ searchPoint0 = {0, 0, 0};
    if (cloud->size() < 50)
        return searchPoint0;

    // compute initial search point as the maximum entropy point
    float max = 0;
    int it;
    for (int i = 0; i < cloud->size(); i++)
    {
        if (cloud->points[i].intensity > max)
        {
            max = cloud->points[i].intensity;
            it = i;
        }
    }
    searchPoint0 = {cloud->points[it].x, cloud->points[it].y, cloud->points[it].z};
    return searchPoint0;
}

void Poly::compute(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &output)
{
    pcl::PointXYZ search0 = initialSearchPoint();

    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> clouds_B;
    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> clouds_F;

    pcl::PointCloud<pcl::PointXYZI>::Ptr source_copy(new pcl::PointCloud<pcl::PointXYZI>);
    pcl::copyPointCloud(*m_source, *source_copy);

    for (int p = 0; p < 3; p++)
    {

        std::cout << "source_copy size: " << source_copy->size() << std::endl;
        std::cout << std::endl;

        std::vector<PolypointsT> pointsVecF, pointsVecB;
        PolypointsT pointPolyF, pointPolyB, pointPolyTempF, pointPolyTempB;

        // INITIAL RUN
        pointPolyF.search = search0;
        pointPolyF.pointParent = search0;
        pointPolyB.search = search0;
        pointPolyB.pointParent = search0;

        computePointsInit(source_copy, pointPolyF, pointPolyB, m_radius, m_step);
        pointsVecF.push_back(pointPolyF);
        pointsVecB.push_back(pointPolyB);
        pointPolyF = {};
        pointPolyB = {};

        // loop forward
        for (int i = 0; i < m_iterations; i++)
        {
            pointPolyTempF = pointsVecF[i];
            if (!pointPolyTempF.pointComputed.getVector3fMap().isZero())
            {
                pointPolyF = {};
                pointPolyF.search = pointPolyTempF.pointComputed;
                pointPolyF.pointParent = pointPolyTempF.centroid;
                if (computeForward(source_copy, pointPolyF, m_radius, m_step) == false)
                    break;
                updatePointCloud(source_copy, pointPolyF.pointParent, m_radius);
                pointsVecF.push_back(pointPolyF);
                pointPolyF = {};
            }
        }

        // loop backward
        for (int i = 0; i < m_iterations; i++)
        {
            pointPolyTempB = pointsVecB[i];
            if (!pointPolyTempB.pointComputed.getVector3fMap().isZero())
            {
                pointPolyB = {};
                pointPolyB.search = pointPolyTempB.pointComputed;
                pointPolyB.pointParent = pointPolyTempB.centroid;
                if (computeBackward(source_copy, pointPolyB, m_radius, m_step) == false)
                    break;
                updatePointCloud(source_copy, pointPolyB.pointParent, m_radius);
                pointsVecB.push_back(pointPolyB);
                pointPolyB = {};
            }
        }

        // put all centroid points inside a new pointcloud
        pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_polyF(new pcl::PointCloud<pcl::PointXYZ>);
        for (int i = 0; i < pointsVecF.size(); i++)
        {
            if (!pointsVecF[i].centroid.getVector3fMap().isZero())
                cloud_polyF->points.push_back(pointsVecF[i].centroid);
        }

        pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_polyB(new pcl::PointCloud<pcl::PointXYZ>);
        for (int i = 0; i < pointsVecB.size(); i++)
        {
            if (!pointsVecB[i].centroid.getVector3fMap().isZero())
                cloud_polyB->points.push_back(pointsVecB[i].centroid);
        }

        clouds_F.push_back(cloud_polyF);
        clouds_B.push_back(cloud_polyB);

        search0 = newSearchPoint(source_copy, cloud_polyF, cloud_polyB);
        if (search0.getVector3fMap().isZero())
            break;
    }

    // prepare output, each pointcloud one curve
    output = prepareOutput(clouds_F, clouds_B);
    visualization(clouds_F, clouds_B, output);
}

void Poly::updatePointCloud(pcl::PointCloud<pcl::PointXYZI>::Ptr &cloud,
                            pcl::PointXYZ &point, float &radius)
{
    std::cout << "cloud initial size: " << cloud->size() << std::endl;

    pcl::PointXYZI searchP;
    searchP.x = point.x;
    searchP.y = point.y;
    searchP.z = point.z;

    pcl::KdTreeFLANN<pcl::PointXYZI> kdtree;
    kdtree.setInputCloud(cloud);
    pcl::Indices indices;
    std::vector<float> sqrDistances;

    if (kdtree.radiusSearch(searchP, radius, indices, sqrDistances) > 0)
    {
        //std::cout << "NN found: " << indices.size() << std::endl;

        pcl::PointIndices::Ptr inliers(new pcl::PointIndices);
        inliers->indices = indices;

        pcl::ExtractIndices<pcl::PointXYZI> extract;
        extract.setInputCloud(cloud);
        extract.setIndices(inliers);
        extract.setNegative(true);
        extract.filter(*cloud);

        //std::cout << "number of points extracted: " << inliers->indices.size() << std::endl;
        //std::cout << "cloud size: " << cloud->size() << std::endl;
    }

    std::cout << "cloud final size: " << cloud->size() << std::endl;
    std::cout << std::endl;
}

void Poly::visualization(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_F,
                         std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_B,
                         std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &output)
{
    // VISUALIZATION
    pcl::visualization::PCLVisualizer viz("PCL Direction Cloud");
    viz.setBackgroundColor(1.0f, 1.0f, 1.0f);
    pcl::visualization::PointCloudColorHandlerGenericField<pcl::PointXYZI> intensity(m_source, "intensity");
    viz.addPointCloud<pcl::PointXYZI>(m_source, intensity, "m_source");
    viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 3, "m_source");
    /*
    for (int k = 0; k < clouds_F.size(); k++)
    {
        viz.addPointCloud<pcl::PointXYZ>(clouds_F[k], "cloud_polyF" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "cloud_polyF" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 1.0f, 0.0f, 0.0f, "cloud_polyF" + std::to_string(k));
        for (int i = 1; i < clouds_F[k]->size(); i++)
        {
            viz.addLine(clouds_F[k]->points[i - 1], clouds_F[k]->points[i], 1.0f, 0.0f, 0.0f, "F" + std::to_string(i) + std::to_string(k));
        }
    }

    for (int k = 0; k < clouds_B.size(); k++)
    {
        viz.addPointCloud<pcl::PointXYZ>(clouds_B[k], "cloud_polyB" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "cloud_polyB" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 0.0f, 0.0f, 1.0f, "cloud_polyB" + std::to_string(k));
        for (int i = 1; i < clouds_B[k]->size(); i++)
        {
            viz.addLine(clouds_B[k]->points[i - 1], clouds_B[k]->points[i], 0.0f, 0.0f, 1.0f, "B" + std::to_string(i) + std::to_string(k));
        }
    }
    */

    for (int k = 0; k < output.size(); k++)
    {
        viz.addPointCloud<pcl::PointXYZ>(output[k], "output" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 5, "output" + std::to_string(k));
        viz.setPointCloudRenderingProperties(pcl::visualization::PCL_VISUALIZER_COLOR, 1.0f, 0.0f, 0.0f, "output" + std::to_string(k));
        for (int i = 1; i < output[k]->size(); i++)
        {
            viz.addLine(output[k]->points[i - 1], output[k]->points[i], 1.0f, 0.0f, 0.0f, "B" + std::to_string(i) + std::to_string(k));
        }
    }
}

std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> Poly::prepareOutput(std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_F,
                                                                     std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> &clouds_B)
{
    std::vector<pcl::PointCloud<pcl::PointXYZ>::Ptr> output;
    if (clouds_B.size() != clouds_F.size())
        PCL_ERROR("Clouds different sizes!!!!!");

    for (int k = 0; k < clouds_B.size(); k++)
    {
        pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
        for (int i = clouds_B[k]->size() - 1; i > 0; i--)
        {
            cloud->points.push_back(clouds_B[k]->points[i]);
        }
        for (int i = 0; i < clouds_F[k]->size(); i++)
        {
            cloud->points.push_back(clouds_F[k]->points[i]);
        }
        output.push_back(cloud);
    }
    return output;
}
